;*****************************************
; Multiply the data with a sine and a cosine function to produce
; an I/Q pair centered at half the Nyqvist frequency.
; Filter the I/Q pair with a simple IIR filter to remove any
; contribution above half the Nyqvist frequency.
; Then skip every second sample.
; Since now no signal is above half the Nyqvist frequency it
; is permitted to "resample" at half the sampling frequency.
; Resampling is just summing points pairwise.
; (An alternative is: skip every second point.)
; By going to I/Q pairs rather than staying with the original data
; the fourier transforms will be more than twice as fast at the expense
; of only a few integer adds and subtracts.
;*************************************
; The IIR filter used attenuates by about 40dB. If there is a strong
; signal in the passband, a "ghost" signal will be present at this level
; (which varies with frequency. 40dB is worst case)
; The ghost signal can be handeled in the last step of the FFT since
; the level and phase is accurately known.
;++++++++++++++
; The IIR filter is described by this C++ program fragment.
; x=input   y=output  z=temporary storage
;    zs=(s1+s3)/2 + s2;
;    s3=s2;
;    s2=s1;
;    s1=(x[i-1]+x[i-2]+x[i-3]-zs)/2
;    y[k]=zs+(s1+s3)/2 + s2
;    s3=s2
;    s2=s1
;    s1=(x[i]+x[i-1]+x[i-2]-y[k])/2;
;    i=i+2
;    k=k+1
; By use of simple coefficients no multiplications are required.
;+++++++++++++++++
; The sine and cosine tables are as follows:
; N   sin    cos
; 0    0      1
; 1   -1      0
; 2    0     -1
; 3    1      0
; Clearly the multiplication by sin and cos is just add and subtract
; if the loop is arranged to take care of 4 data points each time.
;++++++++++++++++++
; Some terms in the FIR filter disappears due to the zeroes in the
; sin/cos tables, so the code does not look much like the explanation
; above. But removing everything we know is zero gives this piece
; of code as a result...

global clear_real2iq
global real_to_complex_input
global compress_rawdat
global expand_rawdat


section .bss
alignb 4
sz3 resb 4 
sz2 resb 4
sz1 resb 4
xs2 resb 4
cz3 resb 4 
cz2 resb 4
cz1 resb 4
xc3 resb 4
rz1 resb 4
xr2 resb 4
rz3 resb 4
rz2 resb 4
bz3 resb 4 
bz2 resb 4
bz1 resb 4
xb3 resb 4
tmp resb 4 
getiq_counter resb 4
start_ptr resb 4



section .text

compress_rawdat:
  push ebx
  push ecx
  push edx
  push esi
  push edi
  mov  esi,[timf1_char]
  mov  edi,[fft1_tmp]
  add  esi,[timf1p_pa]
  mov  ebx,[ad_read_bytes]
; Compress 4 byte words (int) to 18 bit.
cmpr:
  mov  ax,[esi+2]      ;byte 2,3  of word 0
  mov  ch,[esi+1]      ;byte 1 of word 0
  mov  [edi],ax
  shr  ch,2
  mov  ax,[esi+6]      ;byte 2,3  of word 1
  mov  cl,[esi+5]      ;byte 1 of word 1
  mov  [edi+2],ax
  and  cl,11000000B
  or   ch,cl
  shr  ch,2
  mov  ax,[esi+10]      ;byte 2,3  of word 2
  mov  cl,[esi+ 9]      ;byte 1 of word 2
  mov  [edi+4],ax
  and  cl,11000000B
  or   ch,cl
  shr  ch,2
  mov  ax,[esi+14]      ;byte 2,3  of word 3
  mov  cl,[esi+13]      ;byte 1 of word 3
  and  cl,11000000B
  or   ch,cl
  mov  [edi+6],ax
  mov  [edi+8],ch
  add  esi,16
  add  edi,9
  sub  ebx,16
  ja   cmpr
  pop  edi
  pop  esi
  pop  edx
  pop  ecx
  pop  ebx
  ret


expand_rawdat:
  push ebx
  push ecx
  push edx
  push esi
  push edi
  mov  esi,[timf1_char]
  mov  edi,[fft1_tmp]
  add  esi,[timf1p_pa]
  mov  ebx,[ad_read_bytes]
; Expand 18 bit packed data.
expnd:
  mov  eax,[edi-2]    
  mov  ecx,[edi+6]
  xor  cx,cx  
  shr  ecx,2
  mov  ax,cx
  mov  [esi],eax
  mov  eax,[edi]
  xor  cx,cx  
  shr  ecx,2
  mov  ax,cx
  mov  [esi+4],eax
  mov  eax,[edi+2]
  xor  cx,cx  
  shr  ecx,2
  mov  ax,cx
  mov  [esi+8],eax
  mov  eax,[edi+4]
  xor  cx,cx  
  shr  ecx,2
  mov  ax,cx
  mov  [esi+12],eax
  add  esi,16
  add  edi,9
  sub  ebx,16
  ja   expnd
  pop  edi
  pop  esi
  pop  edx
  pop  ecx
  pop  ebx
  ret






clear_real2iq:
  push ebx
  push ecx
  mov  eax,sz3
  mov  ecx,tmp
  sub  ecx,eax
  shr  ecx,2
clrl:  mov dword[eax],0
  add  eax,4    
  loop clrl
  pop  ecx
  pop  ebx
  ret

extern fft1_tmp
extern fft1_char 
extern fft1p_input
extern fft1p_next
extern fft1_bytes
extern fft1_bufmask
extern fft1_save_bytes
extern ad_read_bytes
extern timf1_char
extern ctimf1
extern timf1p_pa
extern ctimf1p_pa
extern ctimf1p_px
extern ctimf1_bytes
extern timf1_bytes
extern timf1p_px
extern timf1_bytemask
extern ctimf1_bytemask
extern getiq_switch
extern fft1_direction

fft1_block_bytes equ ad_read_bytes


real_to_complex_input: 
 push ebx
 push ecx
 push edx
 push esi
 push edi
 push ebp
 mov  esi,[timf1_char]
 mov  edi,[ctimf1]
 add  esi,[timf1p_px]
 add  edi,[ctimf1p_pa]
 mov  ebx,[ad_read_bytes]
 shr  ebx,4
 mov  [getiq_counter],ebx
 mov  eax,[fft1_direction]
 or   eax,eax
 jns  forward
 jmp  real_to_complex_reverse 
forward:
 mov  eax,[getiq_switch]
 or   ah,ah
 jz   sel_dword
 or   al,al
 jz   gosi
 jmp near ss_loop
gosi:
 jmp near si_loop
sel_dword:
 shr  ebx,1
 mov  [getiq_counter],ebx
 or   al,al
 jz   di_loop
 jmp near ds_loop
; Dual channels, input is int
di_loop: mov  ebp,[sz3]
 mov  eax,[sz2]
 mov  edx,[sz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xs2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [edi],ebp
 mov  eax,dword[8+esi]
 sar  eax,4 
 mov [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx 
 sub  ebx,edx
 sar  ebx,1
 mov  [sz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [sz3],eax   ;s3=s2
 add  ebx,edx
 mov  [16+edi],ebx
 mov  eax,dword[24+esi]
 sar  eax,4
 neg  eax
 mov  [xs2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [sz1],eax
;**********
;multiply with cosine table, filter and sum.
 mov  edx,[cz3]
 mov  ebx,[cz2]
 mov  ecx,[cz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 mov  eax,dword[esi]
 sar  eax,4
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xc3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [4+edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 mov  eax,dword[16+esi]
 sar  eax,4
 neg  eax
 mov  [xc3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [cz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [cz3],edx
 add  ebx,ecx
 mov  [20+edi],ebx
 mov  eax,[xc3]
 sub  eax,ebx
 sar  eax,1
 mov  [cz1],eax
;***********************************
 mov  ebp,[rz3]
 mov  eax,[rz2]
 mov  edx,[rz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xr2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [8+edi],ebp
 mov  eax,dword[12+esi]
 sar  eax,4
 mov  [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx
 sub  ebx,edx
 sar  ebx,1
 mov  [rz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [rz3],eax   ;s3=s2
 add  ebx,edx
 mov  [24+edi],ebx
 mov  eax,dword[28+esi]
 sar  eax,4
 neg  eax
 mov  [xr2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [rz1],eax
;**********
;multiply with cosine table, filter and sum.
 mov  edx,[bz3]
 mov  ebx,[bz2]
 mov  ecx,[bz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 mov  eax,dword[4+esi]
 sar  eax,4
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xb3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [12+edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 mov  eax,dword[20+esi]
 sar  eax,4
 neg  eax
 mov  [xb3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [bz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [bz3],edx
 add  ebx,ecx
 mov  [28+edi],ebx
 mov  eax,[xb3]
 sub  eax,ebx
 sar  eax,1
 mov  [bz1],eax
;****************************
 add  esi,32
 add  edi,32
 dec  dword[getiq_counter]
 jnz near di_loop
real_to_complex_x:
 sub  esi,[timf1_char]
 sub  edi,[ctimf1]
 and  esi,[timf1_bytemask]
 and  edi,[ctimf1_bytemask]
 mov  [timf1p_px],esi
 mov  [ctimf1p_pa],edi
 pop ebp
 pop edi
 pop esi
 pop edx
 pop ecx
 pop ebx
 ret



; Dual channels, input is short int
ds_loop: mov  ebp,[sz3]
 mov  eax,[sz2]
 mov  edx,[sz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xs2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [edi],ebp
 movsx eax,word[4+esi]
 mov [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx 
 sub  ebx,edx
 sar  ebx,1
 mov  [sz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [sz3],eax   ;s3=s2
 add  ebx,edx
 mov  [16+edi],ebx
 movsx eax,word[12+esi]
 neg  eax
 mov  [xs2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [sz1],eax
;**********
;multiply with cosine table, filter and sum.
 mov  edx,[cz3]
 mov  ebx,[cz2]
 mov  ecx,[cz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 movsx eax,word[esi]
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xc3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [4+edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 movsx eax,word[8+esi]
 neg  eax
 mov  [xc3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [cz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [cz3],edx
 add  ebx,ecx
 mov  [20+edi],ebx
 mov  eax,[xc3]
 sub  eax,ebx
 sar  eax,1
 mov  [cz1],eax
;***********************************
 mov  ebp,[rz3]
 mov  eax,[rz2]
 mov  edx,[rz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xr2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [8+edi],ebp
 movsx eax,word[6+esi]
 mov  [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx
 sub  ebx,edx
 sar  ebx,1
 mov  [rz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [rz3],eax   ;s3=s2
 add  ebx,edx
 mov  [24+edi],ebx
 movsx eax,word[14+esi]
 neg  eax
 mov  [xr2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [rz1],eax
;**********
;multiply with cosine table, filter and sum.
 mov  edx,[bz3]
 mov  ebx,[bz2]
 mov  ecx,[bz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 movsx eax,word[2+esi]
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xb3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [12+edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 movsx eax,word[10+esi]
 neg  eax
 mov  [xb3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [bz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [bz3],edx
 add  ebx,ecx
 mov  [28+edi],ebx
 mov  eax,[xb3]
 sub  eax,ebx
 sar  eax,1
 mov  [bz1],eax
;****************************
 add  esi,16
 add  edi,32
 dec  dword[getiq_counter]
 jnz near ds_loop
 jmp near real_to_complex_x




; Single channel, input is short int
ss_loop: mov  ebp,[sz3]
 mov  eax,[sz2]
 mov  edx,[sz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xs2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [edi],ebp
 movsx eax,word[2+esi]
 mov [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx 
 sub  ebx,edx
 sar  ebx,1
 mov  [sz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [sz3],eax   ;s3=s2
 add  ebx,edx
 mov  [8+edi],ebx
 movsx eax,word[6+esi]
 neg  eax
 mov  [xs2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [sz1],eax
 mov  edx,[cz3]
 mov  ebx,[cz2]
 mov  ecx,[cz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 movsx eax,word[esi]
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xc3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [4+edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 movsx eax,word[4+esi]
 neg  eax
 mov  [xc3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [cz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [cz3],edx
 add  ebx,ecx
 mov  [12+edi],ebx
 mov  eax,[xc3]
 sub  eax,ebx
 sar  eax,1
 mov  [cz1],eax
 add  esi,8
 add  edi,16
 dec  dword[getiq_counter]
 jnz near ss_loop
 jmp near real_to_complex_x


; Single channel, input is int
si_loop: mov  ebp,[sz3]
 mov  eax,[sz2]
 mov  edx,[sz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xs2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [edi],ebp
 mov  eax,dword[4+esi]
 sar  eax,4
 mov  [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx 
 sub  ebx,edx
 sar  ebx,1
 mov  [sz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [sz3],eax   ;s3=s2
 add  ebx,edx
 mov  [8+edi],ebx
 mov  eax,dword[12+esi]
 sar  eax,4
 neg  eax
 mov  [xs2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [sz1],eax
 mov  edx,[cz3]
 mov  ebx,[cz2]
 mov  ecx,[cz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 mov  eax,dword[esi]
 sar  eax,4
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xc3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [4+edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 mov  eax,dword[8+esi]
 sar  eax,4
 neg  eax
 mov  [xc3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [cz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [cz3],edx
 add  ebx,ecx
 mov  [12+edi],ebx
 mov  eax,[xc3]
 sub  eax,ebx
 sar  eax,1
 mov  [cz1],eax
 add  esi,16
 add  edi,16
 dec  dword[getiq_counter]
 jnz near si_loop
 jmp  real_to_complex_x

real_to_complex_reverse: 
 mov  eax,[getiq_switch]
 or   ah,ah
 jz   sel_dword_rev
 or   al,al
 jz   gosi_rev
 jmp near ss_loop_rev
gosi_rev:
 jmp near si_loop_rev
sel_dword_rev:
 shr  ebx,1
 mov  [getiq_counter],ebx
 or   al,al
 jz   di_loop_rev
 jmp near ds_loop_rev
; Dual channels, input is int
di_loop_rev: mov  ebp,[sz3]
 mov  eax,[sz2]
 mov  edx,[sz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xs2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [4+edi],ebp
 mov  eax,dword[8+esi]
 sar  eax,4 
 mov [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx 
 sub  ebx,edx
 sar  ebx,1
 mov  [sz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [sz3],eax   ;s3=s2
 add  ebx,edx
 mov  [20+edi],ebx
 mov  eax,dword[24+esi]
 sar  eax,4
 neg  eax
 mov  [xs2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [sz1],eax
;**********
;multiply with cosine table, filter and sum.
 mov  edx,[cz3]
 mov  ebx,[cz2]
 mov  ecx,[cz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 mov  eax,dword[esi]
 sar  eax,4
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xc3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 mov  eax,dword[16+esi]
 sar  eax,4
 neg  eax
 mov  [xc3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [cz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [cz3],edx
 add  ebx,ecx
 mov  [16+edi],ebx
 mov  eax,[xc3]
 sub  eax,ebx
 sar  eax,1
 mov  [cz1],eax
;***********************************
 mov  ebp,[rz3]
 mov  eax,[rz2]
 mov  edx,[rz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xr2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [12+edi],ebp
 mov  eax,dword[12+esi]
 sar  eax,4
 mov  [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx
 sub  ebx,edx
 sar  ebx,1
 mov  [rz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [rz3],eax   ;s3=s2
 add  ebx,edx
 mov  [28+edi],ebx
 mov  eax,dword[28+esi]
 sar  eax,4
 neg  eax
 mov  [xr2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [rz1],eax
;**********
;multiply with cosine table, filter and sum.
 mov  edx,[bz3]
 mov  ebx,[bz2]
 mov  ecx,[bz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 mov  eax,dword[4+esi]
 sar  eax,4
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xb3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [8+edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 mov  eax,dword[20+esi]
 sar  eax,4
 neg  eax
 mov  [xb3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [bz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [bz3],edx
 add  ebx,ecx
 mov  [24+edi],ebx
 mov  eax,[xb3]
 sub  eax,ebx
 sar  eax,1
 mov  [bz1],eax
;****************************
 add  esi,32
 add  edi,32
 dec  dword[getiq_counter]
 jnz near di_loop_rev
 jmp real_to_complex_x



; Dual channels, input is short int
ds_loop_rev: mov  ebp,[sz3]
 mov  eax,[sz2]
 mov  edx,[sz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xs2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [4+edi],ebp
 movsx eax,word[4+esi]
 mov [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx 
 sub  ebx,edx
 sar  ebx,1
 mov  [sz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [sz3],eax   ;s3=s2
 add  ebx,edx
 mov  [20+edi],ebx
 movsx eax,word[12+esi]
 neg  eax
 mov  [xs2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [sz1],eax
;**********
;multiply with cosine table, filter and sum.
 mov  edx,[cz3]
 mov  ebx,[cz2]
 mov  ecx,[cz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 movsx eax,word[esi]
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xc3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 movsx eax,word[8+esi]
 neg  eax
 mov  [xc3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [cz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [cz3],edx
 add  ebx,ecx
 mov  [16+edi],ebx
 mov  eax,[xc3]
 sub  eax,ebx
 sar  eax,1
 mov  [cz1],eax
;***********************************
 mov  ebp,[rz3]
 mov  eax,[rz2]
 mov  edx,[rz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xr2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [12+edi],ebp
 movsx eax,word[6+esi]
 mov  [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx
 sub  ebx,edx
 sar  ebx,1
 mov  [rz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [rz3],eax   ;s3=s2
 add  ebx,edx
 mov  [28+edi],ebx
 movsx eax,word[14+esi]
 neg  eax
 mov  [xr2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [rz1],eax
;**********
;multiply with cosine table, filter and sum.
 mov  edx,[bz3]
 mov  ebx,[bz2]
 mov  ecx,[bz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 movsx eax,word[2+esi]
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xb3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [8+edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 movsx eax,word[10+esi]
 neg  eax
 mov  [xb3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [bz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [bz3],edx
 add  ebx,ecx
 mov  [24+edi],ebx
 mov  eax,[xb3]
 sub  eax,ebx
 sar  eax,1
 mov  [bz1],eax
;****************************
 add  esi,16
 add  edi,32
 dec  dword[getiq_counter]
 jnz near ds_loop_rev
 jmp near real_to_complex_x




; Single channel, input is short int
ss_loop_rev: mov  ebp,[sz3]
 mov  eax,[sz2]
 mov  edx,[sz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xs2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [4+edi],ebp
 movsx eax,word[2+esi]
 mov [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx 
 sub  ebx,edx
 sar  ebx,1
 mov  [sz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [sz3],eax   ;s3=s2
 add  ebx,edx
 mov  [12+edi],ebx
 movsx eax,word[6+esi]
 neg  eax
 mov  [xs2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [sz1],eax
 mov  edx,[cz3]
 mov  ebx,[cz2]
 mov  ecx,[cz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 movsx eax,word[esi]
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xc3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 movsx eax,word[4+esi]
 neg  eax
 mov  [xc3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [cz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [cz3],edx
 add  ebx,ecx
 mov  [8+edi],ebx
 mov  eax,[xc3]
 sub  eax,ebx
 sar  eax,1
 mov  [cz1],eax
 add  esi,8
 add  edi,16
 dec  dword[getiq_counter]
 jnz near ss_loop_rev
 jmp near real_to_complex_x


; Single channel, input is int
si_loop_rev: mov  ebp,[sz3]
 mov  eax,[sz2]
 mov  edx,[sz1]
 add  ebp,edx
 sar  ebp,1
 add  ebp,eax
 mov  ecx,[xs2]
 mov  ebx,ecx
 sub  ecx,ebp
 sar  ecx,1
 add  eax,ecx
 sar  eax,1
 add  eax,edx
 add  ebp,eax
 mov  [4+edi],ebp
 mov  eax,dword[4+esi]
 sar  eax,4
 mov  [tmp],eax
 sub  eax,ebp
 add  eax,ebx
 sar  eax,1
 add  edx,eax
 sar  edx,1                ;dx=(s1+s3)/2
 add  edx,ecx              ;dx=(s1+s3)/2 + s2
 mov  ebx,[tmp]
 mov  ebp,ebx 
 sub  ebx,edx
 sar  ebx,1
 mov  [sz2],ebx   ;s2=s1
 add  ebx,ecx
 sar  ebx,1                ;bx=(s1+s3)/2
 add  ebx,eax              ;bx=(s1+s3)/2 + s2
 mov  [sz3],eax   ;s3=s2
 add  ebx,edx
 mov  [12+edi],ebx
 mov  eax,dword[12+esi]
 sar  eax,4
 neg  eax
 mov  [xs2],eax
 sub  eax,ebx
 add  eax,ebp
 sar  eax,1
 mov  [sz1],eax
 mov  edx,[cz3]
 mov  ebx,[cz2]
 mov  ecx,[cz1]
 add  edx,ecx
 sar  edx,1
 add  edx,ebx
 mov  eax,dword[esi]
 sar  eax,4
 mov  [tmp],eax
 sub  eax,edx
 add  eax,[xc3]
 sar  eax,1
 mov  ebp,eax
 add  ebp,ebx
 sar  ebp,1
 add  ebp,ecx
 add  ebp,edx
 mov  [edi],ebp
 mov  ebx,[tmp]
 mov  edx,ebx
 sub  edx,ebp
 sar  edx,1
 add  ecx,edx
 sar  ecx,1
 add  ecx,eax
 mov  ebp,eax
 mov  eax,dword[8+esi]
 sar  eax,4
 neg  eax
 mov  [xc3],eax
 sub  eax,ecx
 add  eax,ebx
 sar  eax,1
 mov  [cz2],eax
 mov  ebx,eax
 add  ebx,ebp
 sar  ebx,1
 add  ebx,edx
 mov  [cz3],edx
 add  ebx,ecx
 mov  [8+edi],ebx
 mov  eax,[xc3]
 sub  eax,ebx
 sar  eax,1
 mov  [cz1],eax
 add  esi,16
 add  edi,16
 dec  dword[getiq_counter]
 jnz near si_loop_rev
 jmp  real_to_complex_x

